#!/usr/bin/python3

"""
  Author:  Yeison Cardona --<yeison.eng@gmail.com>
  Purpose:
  Created: 03/09/15
"""

import inspect
import re
from collections import OrderedDict
from datetime import datetime
import decimal

from django.conf.urls import url, include
from django.http import JsonResponse
from rest_framework.response import Response as RestResponse

from rest_framework import viewsets, status
from rest_framework.reverse import reverse
from rest_framework.views import APIView
from rest_framework import serializers
from rest_framework.utils import formatting
import ast

from django.conf.urls import url

import traceback


from django.core.files.uploadedfile import InMemoryUploadedFile


from .parser import Parser



########################################################################
class DataTypeSerializer(serializers.BaseSerializer):
    """"""

    #----------------------------------------------------------------------
    def to_representation(self, obj):

        return obj


########################################################################
class MethodResponse:
    """"""

    #----------------------------------------------------------------------
    def Response(self, *args, **kwargs):
        """"""
        if "post" in self.ALLOWED:
            return JsonResponse(*args, **kwargs)
        else:
            return RestResponse(*args, **kwargs)




########################################################################
class ModuleViewSet(MethodResponse, viewsets.GenericViewSet):
    """"""
    serializer_class = DataTypeSerializer
    custom_exceptions = tuple()


    #----------------------------------------------------------------------
    def metadata(self, request, method_call):
        """"""
        method= getattr(self.get_module(), method_call)
        argspec = inspect.getargspec(method)

        view_name = self.get_view_name()
        new_view_name = "{}.{}()".format(view_name, method_call)
        setattr(self, "get_view_name", lambda :new_view_name)

        data = OrderedDict()
        data["name"] = method.__name__
        data["args"] = argspec.args[1:]
        if argspec.defaults:
            data["default_args"] = dict(zip(argspec.args[-len(argspec.defaults):], argspec.defaults))


        docstring = inspect.getdoc(method)
        request.META["URL_SCHEME"] = request.META["wsgi.url_scheme"]
        parser = Parser(docstring, request)
        if  parser.examples:
            data["examples"] = parser.examples
        if parser.docstring:
            data["help"] = parser.docstring
        else:
            data["description"] = inspect.getdoc(method)


        return self.Response(data, status=status.HTTP_200_OK)



    #----------------------------------------------------------------------
    def get_queryset(self):
        """"""

    #----------------------------------------------------------------------
    def get_module(self):
        """"""
        if hasattr(self, "module_class"):
            return self.module_class()
        else:
            assert "Must define class."


    #----------------------------------------------------------------------
    def get_python_serializer(self):
        """"""
        return self.serializer_class

    #----------------------------------------------------------------------
    def get_exceptions(self):
        """"""
        return self.custom_exceptions


    #----------------------------------------------------------------------
    def module_handle_exception(self, err):
        """"""
        return err.error


    #----------------------------------------------------------------------
    def pythonize_kwargs(self, kwargs):
        """"""
        regex_listcompresor = "\[[\s]*([\d]*)[\s]*:[\s]*([\d]*)[\s]*\]"

        new_kwargs = {}

        for key in kwargs:
            value = kwargs[key]

            if isinstance(value, InMemoryUploadedFile):
                new_kwargs[key] = value
                continue


            if isinstance(value, (str, bytes)):
                listcompresor = re.match(regex_listcompresor, value)

                if listcompresor:
                    new_kwargs[key] = list(range(int(listcompresor.group(1)), 1+int(listcompresor.group(2))))

                else:
                    # if isinstance(value, list) and len(value) == 1:
                        # try: new_kwargs[key] = ast.literal_eval(value[0])
                        # except: pass
                    # else:
                        # try: new_kwargs[key] = ast.literal_eval(value)
                        # except: pass
                    try:
                        new_kwargs[key] = ast.literal_eval(value)
                    except:
                        new_kwargs[key] = value


        return new_kwargs



    #----------------------------------------------------------------------
    def method_api(self, request, method_call):
        """"""
        if request.method == "GET":
            kwargs = request.query_params
        elif request.method == "POST":
            kwargs = request.data.dict()

        kwargs = self.pythonize_kwargs(kwargs)

        instance = getattr(self, "module_class")
        instance.request = request

        view_name = self.get_view_name()
        new_view_name = "{}.{}()".format(view_name, method_call)
        # setattr(self, "get_view_name", lambda :new_view_name)

        try:
            response = getattr(self.get_module(), method_call)(**kwargs)
            if isinstance(response, dict):
                serializer = self.get_python_serializer()(response, many=False, context={"request": request})
                return self.Response(serializer.data, status=status.HTTP_200_OK)
            elif isinstance(response, (str, list, set, bool, int, float, decimal.Decimal)):
                return self.Response(response, status=status.HTTP_200_OK)
            else:
                return self.Response("ERROR", status=status.HTTP_200_OK)

        except self.get_exceptions() as err:
            return self.Response(self.module_handle_exception(err), status=status.HTTP_400_BAD_REQUEST)

        except (TypeError, KeyError):
            method= getattr(self.get_module(), method_call)
            argspec = inspect.getargspec(method)
            data = OrderedDict()
            data["name"] = method.__name__
            data["args"] = argspec.args[1:]
            if argspec.defaults:
                data["default_args"] = dict(zip(argspec.args[-len(argspec.defaults):], argspec.defaults))
            data["description"] = inspect.getdoc(method)
            data["error"] = traceback.format_exc()
            return self.Response(data, status=status.HTTP_400_BAD_REQUEST)

        data = OrderedDict()
        data["error"] = ":("
        self.Response(data, status=status.HTTP_400_BAD_REQUEST)



    #----------------------------------------------------------------------
    @classmethod
    def get_methods(self):
        """"""
        instance = getattr(self, "module_class")
        methods = list(filter(lambda s:not s.startswith("__"), [method for method in dir(instance)]))

        methods = sorted(methods, key=lambda x:len(x), reverse=True)
        return methods


    #----------------------------------------------------------------------
    @classmethod
    def get_urlpatterns(self, allowed):
        """"""
        methods = self.get_methods()
        urlpatterns = []
        for method in methods:
            urlpatterns.append(url(r"^(?P<method_call>{})/$".format(method), self.as_view(allowed), name="{}".format(method)))

        return urlpatterns


    #----------------------------------------------------------------------
    @classmethod
    def get_urls(self, namespace, allowed):
        """"""

        allowed_ = {"options": "metadata",}
        for key in allowed:
            allowed_[key] = "method_api"

        methods = self.get_methods()
        urlpatterns = self.get_urlpatterns(allowed_)
        instance = self

        urlpatterns.append(url(r"^$", get_APIRoot(methods, namespace, instance).as_view(), name='api-root'),)

        return urlpatterns


    #----------------------------------------------------------------------
    @classmethod
    def include(self, namespace, allowed=["get"]):
        """"""
        self.ALLOWED = allowed
        return include(self.get_urls(namespace=namespace, allowed=allowed), namespace=namespace)



#----------------------------------------------------------------------
def get_APIRoot(methods, namespace, instance):
    """"""
    ########################################################################
    class APIRoot(APIView):
        """"""
        #----------------------------------------------------------------------
        def get(self, request):
            """"""
            response = OrderedDict()
            for method in methods:
                response[method] = reverse("{}:{}".format(namespace, method), args=(method, ), request=request)

            return RestResponse(response)


        #----------------------------------------------------------------------
        def get_view_name(self):
            """"""
            name = instance.__name__
            name = formatting.remove_trailing_string(name, 'View')
            name = formatting.remove_trailing_string(name, 'ViewSet')
            name = formatting.camelcase_to_spaces(name)

            return name




    return APIRoot
